package org.xbib.interlibrary.ui;

import java.io.IOException;
import java.io.OutputStream;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;
import java.util.Locale;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Collectors;
import javax.xml.stream.XMLStreamException;
import org.xbib.datastructures.common.MultiMap;
import org.xbib.datastructures.json.tiny.JsonBuilder;
import org.xbib.datastructures.tiny.TinyMultiMap;
import org.xbib.interlibrary.api.ApplicationModule;
import org.xbib.interlibrary.api.library.Library;
import org.xbib.interlibrary.api.order.Circulation;
import org.xbib.interlibrary.api.order.DefaultStatus;
import org.xbib.interlibrary.api.order.Order;
import org.xbib.interlibrary.api.order.Status;
import org.xbib.interlibrary.api.order.tracking.DisplayStyle;
import org.xbib.interlibrary.api.order.tracking.TrackedOrder;
import org.xbib.interlibrary.api.order.tracking.TrackedOrderList;
import org.xbib.interlibrary.common.Interlibrary;
import org.xbib.interlibrary.common.InterlibraryModule;
import org.xbib.interlibrary.common.order.DefaultTrackedOrder;
import org.xbib.interlibrary.common.order.DefaultTrackedOrderList;
import org.xbib.interlibrary.db.DatabaseModule;

public class ApiModule implements InterlibraryModule {

    private static final Logger logger = Logger.getLogger(ApiModule.class.getName());

    private Interlibrary interlibrary;

    private DatabaseModule databaseModule;

    public ApiModule() {
    }

    @Override
    public String getName() {
        return "ApiModule";
    }

    @Override
    public List<Class<? extends ApplicationModule>> dependsOn() {
        return List.of(DatabaseModule.class);
    }

    @Override
    public void setInterlibrary(Interlibrary interlibrary) {
        this.interlibrary = interlibrary;
        this.databaseModule = interlibrary.getModule(DatabaseModule.class);
    }

    @Override
    public Interlibrary getInterlibrary() {
        return interlibrary;
    }

    @Override
    public void run() {
    }

    public void getOrder(MultiMap<String, Object> params, OutputStream outputStream) throws IOException {
        logger.log(Level.INFO, "executing order method with params = " + params);
        Order order = databaseModule.getOrderManager().getOrder(params.getString("key"));
        if (order != null) {
            outputStream.write(order.toJson().getBytes(StandardCharsets.UTF_8));
        }
    }

    /**
     * Internal Digibib "SOAP XML interface" tracking method.
     *
     * @param params parmas
     * @param outputStream output stream
     * @return true if executed, false if not
     * @throws IOException if execution fails
     * @throws XMLStreamException if XML generation fails
     */
    public boolean executeInterface(MultiMap<String, Object> params,
                                    OutputStream outputStream) throws IOException, XMLStreamException {
        logger.log(Level.INFO, "executing interface method with params = " + params);
        String func = params.getString("Func");
        List<String> isil = new ArrayList<>();
        if (params.containsKey("ISIL")) {
            for (Object object : params.get("ISIL")) {
                isil.add(object.toString());
            }
        }
        String transactionId = params.getString("TransactionId");
        String clientid = params.getString("UserId");
        Integer clientidlength = params.getInteger("UserIdLength", -1);
        EnumSet<Circulation> circulationSet = EnumSet.noneOf(Circulation.class);
        if (params.containsKey("Type")) {
            for (Object object : params.get("Type")) {
                try {
                    String circulationString = object.toString().toUpperCase(Locale.ROOT);
                    switch (circulationString) {
                        case "C", "COPY" -> circulationSet.add(Circulation.COPY);
                        case "L", "LOAN" -> circulationSet.add(Circulation.LOAN);
                        case "LC", "CL", "LOAN,COPY", "COPY,LOAN" -> {
                            circulationSet.add(Circulation.COPY);
                            circulationSet.add(Circulation.LOAN);
                        }
                    }
                } catch (Exception e) {
                    logger.warning("unrecognized circulation: " + object);
                }
            }
        }
        DefaultStatus status = null;
        if (params.containsKey("Status")) {
            for (Object object : params.get("Status")) {
                if (object != null) {
                    try {
                        String statusString = switch (object.toString().toUpperCase(Locale.ROOT)) {
                            case "CANCEL", "CANCELAFL" -> "NOTSUPPLIED";
                            case "EXPIRE" -> "EXPIRE";
                            case "FAILURE" -> "ERROR";
                            case "FINISH" -> "END";
                            case "ZFL2LWS" -> "PREPARING";
                            case "LSW2ZFL", "LWS2ZFL" -> "ROUTING";
                            case "ORDERIN" -> "INCOMING";
                            case "PRINTRL" -> "PROCESSED";
                            case "RETURN" -> "RETURN";
                            case "SHIPPED" -> "SHIP";
                            case "SIGN" -> "SIGN";
                            case "WORKOUT" -> "INPROCESS";
                            case "QUIT" -> "QUIT";
                            default -> throw new IllegalArgumentException("unknown 'Status' parameter: " + object);
                        };
                        status = new DefaultStatus(Status.Code.valueOf(statusString), "","",
                                Status.Substatus.UNDEFINED, "", "");
                    } catch (Exception e) {
                        logger.warning("unrecognized status: " + object);
                    }
                }
            }
        }
        int days = params.getInteger("TimeFrame", 90);
        int size = params.getInteger("ListSize", 10);
        int offset = params.getInteger("Start", 1);
        offset = offset > 0 ? offset - 1 : offset;
        String profileString = params.getString("Profile", "ALL");
        DisplayStyle displayStyle;
        if ("ORDERLIST".equals(func)) {
            displayStyle = switch (profileString.toUpperCase(Locale.ROOT)) {
                case "ALL", "FULL" -> DisplayStyle.INCLUDE_EXTERNAL_TRACKING;
                case "LIBRARY" -> DisplayStyle.SOAP_LIST_LIBRARY;
                default -> DisplayStyle.SOAP_LIST_CLIENT;
            };
            TrackedOrderList trackedOrderList = new DefaultTrackedOrderList();
            TrackedOrderList list = databaseModule.getOrderManager().trackList(isil, clientid, clientidlength,
                    circulationSet, status, days, offset, size, displayStyle);
            logger.info("found " + list.getEntries().size() + " entries in db " + databaseModule);
            trackedOrderList.add(list);
            if (trackedOrderList.isEmpty()) {
                outputStream.write(trackedOrderList.toXml(displayStyle, -1, -1).getBytes(StandardCharsets.UTF_8));
            } else {
                int from = offset + 1;
                int to = from + Math.min(trackedOrderList.getEntries().size(), size) - 1;
                outputStream.write(trackedOrderList.toXml(displayStyle, from, to).getBytes(StandardCharsets.UTF_8));
                return true;
            }
            return true;
        } else if ("DETAIL".equals(func)) {
            displayStyle = switch (profileString.toUpperCase(Locale.ROOT)) {
                case "ALL", "FULL" -> DisplayStyle.INCLUDE_EXTERNAL_TRACKING;
                case "LIBRARY" -> DisplayStyle.SOAP_LIBRARY;
                default -> DisplayStyle.SOAP_CLIENT;
            };
            TrackedOrder trackedOrder = new DefaultTrackedOrder();
            trackedOrder = databaseModule.getOrderManager().track(isil, clientid, clientidlength, transactionId, displayStyle);
            if (trackedOrder.getOrderNumber() != null) {
                outputStream.write(trackedOrder.toXml(displayStyle).getBytes(StandardCharsets.UTF_8));
                return true;
            }
            outputStream.write(trackedOrder.toXml(displayStyle).getBytes(StandardCharsets.UTF_8));
            return true;
        }
        return false;
    }

    /**
     * German ILL CGI order tracking.
     */
    public boolean executeTrack(MultiMap<String, Object> params,
                                OutputStream outputStream) throws IOException, XMLStreamException {
        logger.log(Level.INFO, "executing track method with params = " + params);
        if (!params.containsKey("isil")) {
            throw new IllegalArgumentException("parameter 'isil' missing");
        }
        List<String> isil = params.get("isil").stream().map(Object::toString).collect(Collectors.toList());
        String clientid = params.getString("clientid"); // displayid ???
        String transactionId = params.getString("transactionid");
        TrackedOrder trackedOrder = new DefaultTrackedOrder();
        trackedOrder = databaseModule.getOrderManager().track(isil, clientid, -1, transactionId, DisplayStyle.INCLUDE_EXTERNAL_TRACKING);
        if (trackedOrder.getOrderNumber() != null) {
        }
        outputStream.write(trackedOrder.toXml(DisplayStyle.INCLUDE_EXTERNAL_TRACKING).getBytes(StandardCharsets.UTF_8));
        return true;
    }

    public boolean executeTransportServiceList(MultiMap<String, Object> params,
                                               OutputStream outputStream) throws IOException {
        final MultiMap<String, String> transportServices = new TinyMultiMap<>();
        interlibrary.getConfiguration().getAllLibraryIdentifiers().forEach(isil -> {
            Library library = interlibrary.getConfiguration().getLibrary(isil);
            if (library == null) {
                logger.log(Level.WARNING, "no library found for " + isil);
            } else {
                /*Map<String, Object> map = library.getRaw();
                if (map == null) {
                    logger.log(Level.WARNING, "no raw map for " + isil);
                } else {
                    Object object = map.get("MediaTransport");
                    if (object != null) {
                        List<Map<String, Object>> list = new LinkedList<>();
                        if (object instanceof Map<?, ?>) {
                            list.add((Map<String, Object>) object);
                        } else if (object instanceof List<?>) {
                            list = (List<Map<String, Object>>) object;
                        } else {
                            logger.log(Level.WARNING, "do not know what to do with " + object);
                        }
                        for (Map<String, Object> m : list) {
                            String service = (String) m.get("mediaType");
                            String type = (String) m.get("mediaTransportType");
                            if (type != null) {
                                transportServices.put(type, library.getISIL());
                            }
                        }
                    }
                }*/
            }
        });
        JsonBuilder builder = JsonBuilder.builder();
        builder.beginMap();
        transportServices.asMap().forEach((k, v) -> {
            try {
                builder.buildKey(k);
                builder.buildCollection(v);
            } catch (IOException e) {
                throw new IllegalStateException(e);
            }
        });
        builder.endMap();
        outputStream.write(builder.build().getBytes(StandardCharsets.UTF_8));
        return true;
    }

    @Override
    public void close() throws IOException {

    }
}
