package org.xbib.interlibrary.ui;

import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.interlibrary.action.order.OrderAction;
import org.xbib.interlibrary.action.order.OrderResponse;
import org.xbib.interlibrary.api.action.Action;
import org.xbib.interlibrary.api.action.Response;
import org.xbib.interlibrary.api.action.Service;
import org.xbib.interlibrary.db.DatabaseModule;
import org.xbib.interlibrary.http.server.InterlibraryServer;
import org.xbib.interlibrary.http.server.ServerStatusHandler;
import org.xbib.interlibrary.order.OrderModule;
import org.xbib.interlibrary.ui.admin.UserSwitchAction;
import org.xbib.interlibrary.ui.admin.UserSwitchController;
import org.xbib.interlibrary.ui.dashboard.DashboardController;
import org.xbib.interlibrary.ui.login.LoginAction;
import org.xbib.interlibrary.ui.login.LoginController;
import org.xbib.interlibrary.ui.login.LogoutController;
import org.xbib.interlibrary.ui.theme.ThemeController;
import org.xbib.net.Parameter;
import org.xbib.net.ParameterException;
import org.xbib.net.http.HttpAddress;
import org.xbib.net.http.HttpHeaderNames;
import org.xbib.net.http.HttpHeaderValues;
import org.xbib.net.http.HttpMethod;
import org.xbib.net.http.HttpResponseStatus;
import org.xbib.net.http.j2html.J2HtmlService;
import org.xbib.net.http.server.application.Application;
import org.xbib.net.http.server.auth.FormAuthenticationHandler;
import org.xbib.net.http.server.domain.BaseHttpDomainBuilder;
import org.xbib.net.http.server.domain.BaseHttpSecurityDomain;
import org.xbib.net.http.server.domain.HttpDomain;
import org.xbib.net.http.server.domain.HttpSecurityDomain;
import org.xbib.net.http.server.handler.InternalServerErrorHandler;
import org.xbib.net.http.server.ldap.LdapContextFactory;
import org.xbib.net.http.server.ldap.LdapGroupMapping;
import org.xbib.net.http.server.ldap.LdapRealm;
import org.xbib.net.http.server.ldap.LdapUserMapping;
import org.xbib.net.http.server.resource.ClassLoaderResourceHandler;
import org.xbib.net.http.server.route.BaseHttpRouterBuilder;
import org.xbib.net.http.server.service.BaseHttpService;
import org.xbib.net.mime.stream.Hex;
import org.xbib.settings.Settings;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;

import static org.xbib.net.http.HttpHeaderNames.CONTENT_LENGTH;
import static org.xbib.net.http.HttpHeaderNames.CONTENT_TYPE;

public final class HttpUIServer extends InterlibraryServer {

    private static final Logger logger = Logger.getLogger(HttpUIServer.class.getName());

    public HttpUIServer() {
        super();
    }

    @Override
    protected void customizeRouter(BaseHttpRouterBuilder baseHttpRouterBuilder) {
        baseHttpRouterBuilder
                .setHandler(400, new ServerStatusHandler(HttpResponseStatus.BAD_REQUEST))
                .setHandler(401, new ServerStatusHandler(HttpResponseStatus.UNAUTHORIZED))
                .setHandler(403, new ServerStatusHandler(HttpResponseStatus.FORBIDDEN))
                .setHandler(404, new ServerStatusHandler(HttpResponseStatus.NOT_FOUND))
                .setHandler(500, new InternalServerErrorHandler());
    }

    @Override
    protected HttpDomain buildDomain(HttpAddress httpAddress,
                                     HttpSecurityDomain securityDomain,
                                     BaseHttpDomainBuilder domainBuilder) throws IOException {
        String prefix = Objects.requireNonNull(getSettings().get("web.prefix"), "web.prefix must not be null");

        ApiModule apiModule = Objects.requireNonNull(getModule(ApiModule.class));
        OrderModule orderModule = Objects.requireNonNull(getModule(OrderModule.class));

        LoginAction loginAction = new LoginAction();
        LoginController loginController = new LoginController();
        LogoutController logoutController = new LogoutController();
        ThemeController themeController = new ThemeController();
        UserSwitchAction userSwitchAction = new UserSwitchAction();
        UserSwitchController userSwitchController = new UserSwitchController();
        DashboardController dashboardController = new DashboardController();

        return domainBuilder
            .setHttpAddress(buildHttpAddress())
            .addService(BaseHttpService.builder()
                    .setPath("/favicon.ico")
                    .setHandler(ctx -> {
                        ctx.status(HttpResponseStatus.OK)
                                .setHeader(HttpHeaderNames.CONTENT_TYPE, "image/x-icon")
                                .body(Hex.fromHex(hexFavIcon));
                        ctx.done();
                    })
                    .build())
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/resources/webjars/**")
                    .setHandler(new ClassLoaderResourceHandler(HttpUIServer.class.getClassLoader(), "META-INF/", 24 * 3600))
                    .build())
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/resources/dist/**")
                    .setHandler(new ClassLoaderResourceHandler(HttpUIServer.class.getClassLoader(), "META-INF/", 24 * 3600))
                    .build())
            .addService(J2HtmlService.builder()
                    .setPrefix(prefix)
                    .setPath("/action/login/{loginform}")
                    .setMethod(HttpMethod.POST)
                    .setSecurityDomain(securityDomain)
                    .setHandler(loginAction)
                    .build())
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/action/theme/{theme}")
                    .setSecurityDomain(securityDomain)
                    .setHandler(themeController)
                    .build())
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/action/userswitch")
                    .setSecurityDomain(securityDomain)
                    .setHandler(userSwitchAction)
                    .build())
            .addService(J2HtmlService.builder()
                    .setPrefix(prefix)
                    .setPath("/login")
                    .setHandler(loginController)
                    .build())
            .addService(J2HtmlService.builder()
                    .setPrefix(prefix)
                    .setPath("/logout")
                    .setSecurityDomain(securityDomain)
                    .setHandler(logoutController)
                    .build())
            .addService(J2HtmlService.builder()
                    .setPrefix(prefix)
                    .setPath("/admin/userswitch/{userswitchform}")
                    .setSecurityDomain(securityDomain)
                    .setHandler(userSwitchController)
                    .build())
            .addService(J2HtmlService.builder()
                    .setPrefix(prefix)
                    .setPath("/admin/userswitch")
                    .setSecurityDomain(securityDomain)
                    .setHandler(userSwitchController)
                    .build())
            .addService(J2HtmlService.builder()
                    .setPrefix(prefix)
                    .setPath("/dashboard")
                    .setSecurityDomain(securityDomain)
                    .setHandler(dashboardController)
                    .build())
            .addService(BaseHttpService.builder()
                    .setPath("/getorder/**")
                    .setMethod(HttpMethod.POST, HttpMethod.GET)
                    .setHandler(ctx -> {
                        try {
                            String contentType = ctx.getRequest().getHeaders().get(CONTENT_TYPE);
                            if (contentType != null && contentType.startsWith("application/json")) {
                                ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
                                apiModule.getOrder(ctx.getRequest().getParameter().asMultiMap(), outputStream);
                                ctx.status(HttpResponseStatus.OK)
                                        .setHeader(CONTENT_TYPE, "application/json")
                                        .setHeader(CONTENT_LENGTH, Long.toString(outputStream.size()))
                                        .body(new ByteArrayInputStream(outputStream.toByteArray()), 8192);
                            } else {
                                ctx.status(HttpResponseStatus.BAD_REQUEST);
                            }
                        } catch (Exception e) {
                            logger.log(Level.SEVERE, e.getMessage(), e);
                            ctx.status(HttpResponseStatus.INTERNAL_SERVER_ERROR)
                                    .setHeader(CONTENT_TYPE, "text/plain; charset=UTF-8")
                                    .body(e.getMessage())
                                    .done();
                        }
                    })
                    .build())
            .addService(BaseHttpService.builder()
                    .setPath("/interface/**")
                    .setMethod(HttpMethod.POST, HttpMethod.GET)
                    .setHandler(ctx -> {
                        try {
                            ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
                            if (apiModule.executeInterface(ctx.getRequest().getParameter().asMultiMap(), outputStream)) {
                                ctx.status(HttpResponseStatus.OK)
                                        .setHeader(CONTENT_TYPE, "application/xml; charset=UTF-8")
                                        .setHeader(CONTENT_LENGTH, Long.toString(outputStream.size()))
                                        .body(new ByteArrayInputStream(outputStream.toByteArray()), 8192);
                            } else {
                                ctx.status(HttpResponseStatus.NOT_FOUND);
                            }
                        } catch (Exception e) {
                            logger.log(Level.SEVERE, e.getMessage(), e);
                            ctx.status(HttpResponseStatus.INTERNAL_SERVER_ERROR)
                                    .setHeader(CONTENT_TYPE, "text/plain; charset=UTF-8")
                                    .body(e.getMessage())
                                    .done();
                        }
                    })
                    .build())
            .addService(BaseHttpService.builder()
                    .setPath("/track/**")
                    .setMethod(HttpMethod.POST, HttpMethod.GET)
                    .setHandler(ctx -> {
                        try {
                            ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
                            if (apiModule.executeTrack(ctx.getRequest().getParameter().asMultiMap(), outputStream)) {
                                ctx.status(HttpResponseStatus.OK)
                                        .setHeader(CONTENT_TYPE, "application/html; charset=UTF-8")
                                        .setHeader(CONTENT_LENGTH, Long.toString(outputStream.size()))
                                        .body(new ByteArrayInputStream(outputStream.toByteArray()), 8192);
                            } else {
                                ctx.status(HttpResponseStatus.NOT_FOUND);
                            }
                        } catch (Exception e) {
                            logger.log(Level.SEVERE, e.getMessage(), e);
                            ctx.status(HttpResponseStatus.INTERNAL_SERVER_ERROR)
                                    .setHeader(CONTENT_TYPE, "text/plain; charset=UTF-8")
                                    .body(e.getMessage())
                                    .done();
                        }
                    })
                    .build())
        .addService(BaseHttpService.builder()
                .setPath("/status")
                .setHandler(ctx -> {
                    ctx.status(HttpResponseStatus.OK)
                            .setHeader(CONTENT_TYPE, HttpHeaderValues.TEXT_PLAIN)
                            .charset(StandardCharsets.UTF_8)
                            .body(getActionStates())
                            .done();
                })
                .build())
        .addService(BaseHttpService.builder()
                .setPath("/service/{service}/{status}")
                .setMethod(HttpMethod.PUT)
                .setHandler(ctx -> {
                    Parameter parameter = ctx.getRequest().getParameter();
                    try {
                        Object serviceName = parameter.get("service", Parameter.Domain.PATH);
                        Object serviceStatus = parameter.get("status", Parameter.Domain.PATH);
                        if (serviceName != null && serviceStatus != null) {
                            if ("true".equals(serviceStatus) || "false".equals(serviceStatus)) {
                                setActionService(OrderAction.INSTANCE, serviceName.toString(), Boolean.parseBoolean(serviceStatus.toString()));
                            }
                            ctx.status(HttpResponseStatus.OK)
                                    .setHeader(CONTENT_TYPE, HttpHeaderValues.TEXT_PLAIN)
                                    .charset(StandardCharsets.UTF_8)
                                    .body("OK");
                        } else {
                            ctx.status(HttpResponseStatus.BAD_REQUEST);
                        }
                    } catch (ParameterException e) {
                        logger.log(Level.SEVERE, e.getMessage(), e);
                        ctx.status(HttpResponseStatus.BAD_REQUEST);
                    }
                })
                .build())
        .addService(BaseHttpService.builder()
                .setPath("/transport/{ordersystem}/{ordernumber}")
                .setMethod(HttpMethod.GET, HttpMethod.POST)
                .setHandler(ctx -> {
                    try {
                        OrderResponse orderResponse = orderModule.transport(ctx.getRequest().getParameter());
                        if (orderResponse.getStatus() == Response.Status.ERROR) {
                            ctx.status(HttpResponseStatus.INTERNAL_SERVER_ERROR)
                                    .setHeader(CONTENT_TYPE, HttpHeaderValues.APPLICATION_JSON)
                                    .body(orderResponse.asJson());
                        } else {
                            ctx.status(HttpResponseStatus.OK)
                                    .setHeader(CONTENT_TYPE, HttpHeaderValues.APPLICATION_JSON)
                                    .body(orderResponse.asJson());
                        }
                        orderResponse.log();
                    } catch (ParameterException e) {
                        logger.log(Level.SEVERE, e.getMessage(), e);
                        ctx.status(HttpResponseStatus.BAD_REQUEST);
                    }
                })
                .build())
            .build();
    }

    @Override
    protected void customizeApplication(Application application) {
        DatabaseModule databaseModule = getModule(DatabaseModule.class);
        application.getAttributes().put("database", databaseModule);
        application.addModule(new UIApplicationModule(this, application, "app", getSettings()));
    }

    @Override
    protected HttpSecurityDomain buildSecurityDomain() {
        Settings settings = getSettings();
        Map<String, LdapContextFactory> contextFactories = new HashMap<>();
        LdapContextFactory contextFactory = new LdapContextFactory(settings.get("ldap.context.authentication", "simple"),
                settings.get("ldap.context.factory", "com.sun.jndi.ldap.LdapCtxFactory"),
                settings.get("ldap.context.realm", null),
                settings.get("ldap.url"),
                settings.getAsBoolean("ldap.context.starttls", false),
                settings.get("ldap.user"),
                settings.get("ldap.password"),
                settings.get("ldap.context.referral", "follow")
        );
        contextFactories.put("default", contextFactory);
        Map<String, LdapUserMapping> userMappings = new HashMap<>();
        LdapUserMapping userMapping = new LdapUserMapping(settings.get("ldap.usermapping.basedn"),
                settings.get("ldap.usermapping.filter"),
                settings.get("ldap.usermapping.uidattribute"),
                settings.get("ldap.usermapping.realnameattribute")
        );
        userMappings.put("default", userMapping);
        Map<String, LdapGroupMapping> groupMappings = new HashMap<>();
        LdapGroupMapping groupMapping = new LdapGroupMapping(settings.get("ldap.groupmapping.basedn"),
                settings.get("ldap.groupmapping.idattribute"),
                settings.get("ldap.groupmapping.filter"),
                settings.getAsArray("ldap.groupmapping.filterargnames")
        );
        groupMappings.put("default", groupMapping);
        LdapRealm ldapRealm = new LdapRealm(settings.get("ldap.realm.name", "[ldap.realm.name undefined]"),
                contextFactories, userMappings, groupMappings);
        FormAuthenticationHandler formAuthenticationHandler =
                new FormAuthenticationHandler("username", "password", ldapRealm);
        return BaseHttpSecurityDomain.builder()
                .setSecurityRealm(ldapRealm)
                .setHandlers(formAuthenticationHandler)
                .build();
    }

    private String getActionStates() {
        StringBuilder sb = new StringBuilder();
        sb.append("Actions\n-------\n");
        for (Action<?, ?, ?, ?> action : getActions()) {
            sb.append(action.getInstance().getClass().getName()).append("\n");
            for (Service<?, ?, ?> service : action.getInstance().getServices()) {
                sb.append("  ").append(service.getName())
                        .append("  ").append(service.isEnabled());
                if (service.getLastThrowable() != null) {
                    sb.append("  ").append(service.getLastThrowable().getMessage());
                }
                sb.append("\n");
            }
        }
        return sb.toString();
    }
}
