package org.xbib.interlibrary.ui.statistics;

import org.xbib.datastructures.tiny.TinyMap;
import org.xbib.event.common.EventManager;
import org.xbib.interlibrary.api.library.Role;
import org.xbib.interlibrary.db.DatabaseModule;
import org.xbib.interlibrary.http.server.BaseInterlibraryServerModule;
import org.xbib.interlibrary.http.server.InterlibraryServer;
import org.xbib.interlibrary.statistics.RoyaltiesStatistics;
import org.xbib.interlibrary.statistics.StatisticsModule;
import org.xbib.interlibrary.ui.statistics.event.RoyaltiesStatisticsClockEvent;
import org.xbib.interlibrary.ui.statistics.event.RoyaltiesStatisticsClockEventConsumer;
import org.xbib.net.Parameter;
import org.xbib.net.UserProfile;
import org.xbib.net.http.HttpHeaderNames;
import org.xbib.net.http.HttpResponseStatus;
import org.xbib.net.http.j2html.J2HtmlService;
import org.xbib.net.http.server.HttpRequest;
import org.xbib.net.http.server.domain.HttpDomainBuilder;
import org.xbib.net.http.server.domain.HttpSecurityDomain;
import org.xbib.net.http.server.route.HttpRouterContext;
import org.xbib.net.http.server.service.BaseHttpService;
import org.xbib.net.http.server.session.Session;
import org.xbib.settings.Settings;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.Month;
import java.time.Year;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Objects;
import java.util.logging.Level;
import java.util.logging.Logger;

public class RoyaltiesModule extends BaseInterlibraryServerModule {

    private static final Logger logger = Logger.getLogger(RoyaltiesModule.class.getName());

    private InterlibraryServer server;

    public RoyaltiesModule() {
    }

    @Override
    public String getName() {
        return "royalties";
    }

    public void installEventServices(EventManager.EventManagerBuilder builder) {
        RoyaltiesStatisticsClockEventConsumer royaltiesStatisticsClockEventConsumer = new RoyaltiesStatisticsClockEventConsumer(this);
        builder
                .register("royalties-clock-event", RoyaltiesStatisticsClockEvent.class)
                .register(royaltiesStatisticsClockEventConsumer);
    }

    @Override
    public void installInto(InterlibraryServer interlibraryServer,
                            HttpSecurityDomain securityDomain,
                            HttpDomainBuilder domainBuilder) {
        this.server = interlibraryServer;
        Settings settings = interlibraryServer.getSettings();
        String prefix = Objects.requireNonNull(settings.get("web.prefix"), "web.prefix must not be null");
        RoyaltiesController royaltiesController = new RoyaltiesController();
        RoyaltiesAction royaltiesAction = new RoyaltiesAction();
        RoyaltiesDownloadAction royaltiesDownloadAction = new RoyaltiesDownloadAction();
        RoyaltiesChangeAction royaltiesChangeAction = new RoyaltiesChangeAction();
        RoyaltiesSearchAction royaltiesSearchAction = new RoyaltiesSearchAction();
        RoyaltiesClockAction royaltiesClockAction = new RoyaltiesClockAction();
        domainBuilder
                .addService(BaseHttpService.builder()
                        .setPrefix(prefix)
                        .setPath("/action/royalties/change/{number}/{value}")
                        .setSecurityDomain(securityDomain)
                        .setHandler(royaltiesChangeAction)
                        .build())
                .addService(BaseHttpService.builder()
                        .setPrefix(prefix)
                        .setPath("/action/royalties/download/{suffix}")
                        .setSecurityDomain(securityDomain)
                        .setHandler(royaltiesDownloadAction)
                        .build())
                .addService(BaseHttpService.builder()
                        .setPrefix(prefix)
                        .setPath("/action/royalties/search/{royaltiesform}")
                        .setSecurityDomain(securityDomain)
                        .setHandler(royaltiesSearchAction)
                        .build())
                .addService(BaseHttpService.builder()
                        .setPrefix(prefix)
                        .setPath("/action/royalties/{royaltiesform}/{key}/{value}")
                        .setSecurityDomain(securityDomain)
                        .setHandler(royaltiesAction)
                        .build())
                .addService(BaseHttpService.builder()
                        .setPrefix(prefix)
                        .setPath("/action/royalties/clock/{startdate}/{enddate}")
                        .setHandler(royaltiesClockAction)
                        .build())
                .addService(BaseHttpService.builder()
                        .setPrefix(prefix)
                        .setPath("/action/royalties")
                        .setSecurityDomain(securityDomain)
                        .setHandler(royaltiesAction)
                        .build())
                .addService(J2HtmlService.builder()
                        .setPrefix(prefix)
                        .setPath("/royalties/{royaltiesform}")
                        .setSecurityDomain(securityDomain)
                        .setHandler(royaltiesController)
                        .build())
                .addService(J2HtmlService.builder()
                        .setPrefix(prefix)
                        .setPath("/royalties")
                        .setSecurityDomain(securityDomain)
                        .setHandler(royaltiesController)
                        .build());
    }

    @SuppressWarnings("unchecked")
    public Map<String, Object> getRoyaltiesForm(HttpRequest request,
                                                Session session) {
        StatisticsModule statisticsModule = Objects.requireNonNull(server.getModule(StatisticsModule.class),
                "statistics module must be present");
        Role role = Role.REQUESTER;
        UserProfile userProfile = request.getContext().getAttributes().get(UserProfile.class, "userprofile");
        String isil = userProfile.getEffectiveRoles().contains("Admin") ? null : userProfile.getEffectiveUserId();
        List<String> halves = (List<String>) session.get("_royalties_halves");
        if (halves == null) {
            halves = statisticsModule.getRoyaltiesPeriods(role, isil);
            if (halves != null && !halves.isEmpty()) {
                session.put("_royalties_halves", halves);
            }
        }
        if (halves == null || halves.isEmpty()) {
            // current date as fallback if statistics is empty
            halves = List.of(Year.now() + "-" + (LocalDate.now().getMonthValue() < 7 ? "1" : "2"));
        }
        TinyMap.Builder<String, Object> builder = RoyaltiesModel.getForm(request.getParameter(), session, halves.getLast());
        builder.put("isil", isil);
        builder.put("periods", halves);
        Map<String, Object> map = builder.build();
        String id = (String) map.get("_id");
        if (id == null) {
            logger.log(Level.WARNING, "no ID found for session update, map = " + map);
            return map;
        }
        session.put(id, map);
        return map;
    }


    public void executeRoyalties(HttpRouterContext context) throws IOException {
        HttpRequest request = context.getRequest();
        Session session = context.getAttributes().get(Session.class, "session");
        Map<String, Object> map = Objects.requireNonNull(getRoyaltiesForm(request, session));
        // right now, we do not use offset, size, page for royalties generation. Maybe later.
        UserProfile userProfile = context.getAttributes().get(UserProfile.class, "userprofile");
        StatisticsModule statisticsModule =
                Objects.requireNonNull(server.getModule(StatisticsModule.class), "statistics module must be present");
        Role role = Role.REQUESTER; // we use the requester role always by KOBV directive
        String isil = userProfile.getEffectiveRoles().contains("Admin") ? null : userProfile.getEffectiveUserId();
        String consortium = "HBZ";
        String period = (String) map.get("period");
        // derive start and end from period
        int year = Integer.parseInt(period.substring(0, 4));
        Month startMonth = period.endsWith("1") ? Month.JANUARY : Month.JULY;
        LocalDateTime start = LocalDateTime.of(year, startMonth, 1, 0, 0, 0);
        LocalDateTime end = start.plusMonths(6L);
        StatisticsModule.Target target = StatisticsModule.Target.READ;
        Path path = null;
        logger.log(Level.INFO, "creating royalties table for period " + period + " isil = " + isil);
        RoyaltiesStatistics statistics = statisticsModule.getRoyalties(role, isil, consortium, period,
                start, end, target, path);
        session.put("_statistics_royalties", statistics.asMap());
    }

    public void executeRoyaltiesDownload(HttpRouterContext context,
                                         StatisticsModule.Target target,
                                         String contentType) throws IOException {
        HttpRequest request = context.getRequest();
        Session session = context.getAttributes().get(Session.class, "session");
        Map<String, Object> map = Objects.requireNonNull(getRoyaltiesForm(request, session));
        UserProfile userProfile = context.getAttributes().get(UserProfile.class, "userprofile");
        StatisticsModule statisticsModule =
                Objects.requireNonNull(server.getModule(StatisticsModule.class), "statistics module must be present");
        Role role = Role.REQUESTER;
        String isil = userProfile.getEffectiveRoles().contains("Admin") ? null : userProfile.getEffectiveUserId();
        String consortium = "HBZ";
        String period = Objects.requireNonNull((String) map.get("period"), "period required");
        // derive start and end from period
        int year = Integer.parseInt(period.substring(0, 4));
        Month startMonth = period.endsWith("1") ? Month.JANUARY : Month.JULY;
        LocalDateTime start = LocalDateTime.of(year, startMonth, 1, 0, 0, 0);
        LocalDateTime end = start.plusMonths(6L);
        String suffix = target.name().toLowerCase(Locale.ROOT);
        Path path = Files.createTempFile("tantiemen_", suffix);
        try {
            RoyaltiesStatistics statistics = statisticsModule.getRoyalties(role, isil, consortium, period,
                    start, end, target, path);
            String fileName = "tantiemen_" + (isil != null ? isil + "_" : "") + period + "." + suffix;
            byte[] b = Files.readAllBytes(path);
            logger.log(Level.INFO, "creating download for period " + period + " isil = " + isil +
                    " suffix = " + suffix + " size = " + b.length);
            context.status(HttpResponseStatus.OK)
                    .setHeader(HttpHeaderNames.CONTENT_TYPE, contentType)
                    .setHeader(HttpHeaderNames.CACHE_CONTROL, "no-cache")
                    .setHeader(HttpHeaderNames.CONTENT_DISPOSITION, "attachment; filename=" + fileName)
                    .body(b)
                    .done();
        } catch (IOException e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
            context.status(HttpResponseStatus.INTERNAL_SERVER_ERROR)
                    .done();
        } finally {
            Files.delete(path);
        }
    }

    public void executeRoyaltiesChange(String number, String value) {
        StatisticsModule statisticsModule =
                Objects.requireNonNull(server.getModule(StatisticsModule.class), "statistics module must be present");
        statisticsModule.changeRoyalties(Role.REQUESTER, number, value);
    }

    public void executeRoyalties(RoyaltiesStatisticsClockEvent event) throws IOException {
        logger.log(Level.INFO, "executing royalties statistics generation, triggered by event " + event);
        DatabaseModule databaseModule =
                Objects.requireNonNull(server.getModule(DatabaseModule.class), "database must be present");
        LocalDateTime end = LocalDateTime.ofInstant(event.getInstant(), ZoneId.systemDefault());
        LocalDateTime start = end.minusMonths(6L);
        databaseModule.getStatisticsManager().putRoyaltiesStatistics(Role.REQUESTER, start, end);
    }

    public void executeRoyaltiesClock(LocalDateTime start, LocalDateTime end) throws IOException {
        DatabaseModule databaseModule =
                Objects.requireNonNull(server.getModule(DatabaseModule.class), "database must be present");
        databaseModule.getStatisticsManager().putRoyaltiesStatistics(Role.REQUESTER, start, end);
    }

    // api call without form
    public void executeRoyaltiesApi(HttpRouterContext context) {
        StatisticsModule statisticsModule =
                Objects.requireNonNull(server.getModule(StatisticsModule.class), "statistics module must be present");
        Settings settings = server.getSettings();
        try {
            String roleString = context.getRequest().getParameter().getAsString("role", Parameter.Domain.PATH);
            if (!"requester".equalsIgnoreCase(roleString) && !"supplier".equalsIgnoreCase(roleString)) {
                context.status(HttpResponseStatus.BAD_REQUEST);
                return;
            }
            roleString = roleString.toLowerCase(Locale.ROOT);
            Role role = Role.valueOf(roleString.toUpperCase(Locale.ROOT));
            String targetString = context.getRequest().getParameter().getAsString("target", Parameter.Domain.PATH);
            if (!"tsv".equalsIgnoreCase(targetString) && !"xml".equalsIgnoreCase(targetString)) {
                context.status(HttpResponseStatus.BAD_REQUEST);
                return;
            }
            targetString = targetString.toLowerCase(Locale.ROOT);
            StatisticsModule.Target target = StatisticsModule.Target.valueOf(targetString.toUpperCase(Locale.ROOT));
            String startParameter = context.getRequest().getParameter().getAsString("start", Parameter.Domain.QUERY);
            if (startParameter == null || startParameter.isEmpty()) {
                context.status(HttpResponseStatus.BAD_REQUEST);
                return;
            }
            String endParameter = context.getRequest().getParameter().getAsString("end", Parameter.Domain.QUERY);
            if (endParameter == null || endParameter.isEmpty()) {
                context.status(HttpResponseStatus.BAD_REQUEST);
                return;
            }
            DateTimeFormatter formatter = DateTimeFormatter.ofPattern("yyyy-MM-dd");
            LocalDate startLocalDate = LocalDate.parse(startParameter, formatter);
            LocalDateTime start = startLocalDate.atStartOfDay();
            LocalDate endLocalDate = LocalDate.parse(endParameter, formatter);
            LocalDateTime end = endLocalDate.atStartOfDay();
            int year = start.getYear();
            String consortium = settings.get("royalties." + roleString + ".consortium", "HBZ");
            StringBuilder period = new StringBuilder(Integer.toString(year));
            LocalDate firstOfJuly = LocalDate.of(start.getYear(), Month.JULY.getValue(), 1);
            if (startLocalDate.isAfter(firstOfJuly)) {
                period.append("-2");
            } else {
                period.append("-1");
            }
            String statisticsMarker = settings.get("royalties." + roleString + ".marker", "tantiemen" +
                    "_" + ("requester".equals(roleString) ? "PFL" : "AFL"));
            String base = settings.get("royalties." + roleString + ".path",
                    "/data/medea/stat/tantiemen/" + Year.now() + "/");
            Path path = target == StatisticsModule.Target.GENERATE ? null :
                    Paths.get(base, Integer.toString(year), roleString,
                            statisticsMarker + "_" + start.format(formatter) + "_" + end.format(formatter) + "." + targetString);
            logger.log(Level.INFO, "start of generating statistics at path " + path);
            statisticsModule.getRoyalties(role, null, consortium, period.toString(), start, end, target, path);
            logger.log(Level.INFO, "end of generating statistics at path " + path);
            context.status(HttpResponseStatus.OK);
        } catch (Exception e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
            context.status(HttpResponseStatus.BAD_REQUEST);
        }
    }
}
